#include <avr\io.h>
#define F_CPU 16000000
#include <util\delay.h>
#include <stdio.h>
#include <avr\interrupt.h>

//////////////////////////////////////////////////////////
//                  Motor defines                       //
//////////////////////////////////////////////////////////
// Canale pentru cei trei senzori
#define FRONT_CHANNEL 0
#define RIGHT_CHANNEL 1
#define BACK_CHANNEL  2
#define TAHO_CHANNEL  4

#define FRONT 0
#define RIGHT 1
#define BACK  2

uint8_t channels[] = {FRONT_CHANNEL, RIGHT_CHANNEL, BACK_CHANNEL};

volatile uint8_t channel = FRONT;

volatile int voltages[] = {0, 0, 0};

// Starea in care se afla cele doua motoare
#define GOING_FORWARD  0
#define GOING_BACKWARD 1
#define STAYING        2

#define TURNING_LEFT  0
#define TURNING_RIGHT 1
#define NOT_TURNING   2

#define UNINITIALIZED  3


volatile int front_back_status = UNINITIALIZED;
volatile int left_right_status = UNINITIALIZED;


// Porturile pe care comand motoarele
#define BACK_MOTOR_IN_1  PC0
#define BACK_MOTOR_IN_2  PC2
#define BACK_MOTOR_EN    PD7

#define FRONT_MOTOR_IN_1  PD4
#define FRONT_MOTOR_IN_2  PD5
#define FRONT_MOTOR_EN    PB3

// Actiunile efectuate de masina

// Motorul din spate
#define STAY       do { if (front_back_status == STAYING) break; \
                        PORTC &= ~((1<<BACK_MOTOR_IN_1) | (1<<BACK_MOTOR_IN_2)); \
						front_back_status = STAYING; } while(0)
#define GO_FORWARD do { if (front_back_status == GOING_FORWARD) break; \
                        STAY; _delay_ms(2); PORTC |= 1<<BACK_MOTOR_IN_2;    PORTC &= ~(1<<BACK_MOTOR_IN_1); \
						front_back_status = GOING_FORWARD; } while(0)
#define GO_BACK    do { if (front_back_status == GOING_BACKWARD) break; \
                        STAY; _delay_ms(2); PORTC = (PORTC | (1<<BACK_MOTOR_IN_1)) & (~(1<<BACK_MOTOR_IN_2)); \
						front_back_status = GOING_BACKWARD; } while(0)

// Motorul din fata
#define GO_STRAIGHT do { if (left_right_status == NOT_TURNING) break; \
                         PORTD &= ~(1<<FRONT_MOTOR_IN_1); PORTD &= ~(1<<FRONT_MOTOR_IN_2); \
						 left_right_status = NOT_TURNING; } while(0)
#define TURN_LEFT   do { if (left_right_status == TURNING_LEFT) break; \
                         \
						 PORTD = (PORTD | (1<<FRONT_MOTOR_IN_1)) & ~(1<<FRONT_MOTOR_IN_2);\
						 left_right_status = TURNING_LEFT; } while(0)
#define TURN_RIGHT  do { if (left_right_status == TURNING_RIGHT) break; \
                         \
						 PORTD = (PORTD | (1<<FRONT_MOTOR_IN_2)) & ~(1<<FRONT_MOTOR_IN_1);\
						 left_right_status = TURNING_RIGHT; } while(0)

// Viteza cu care merge motorul din fata in spate/fata
#define DEFAULT_BACK_OCR 	130
#define DEFAULT_FRONT_OCR   130

//////////////////////////////////////////////////////////
//                  Sensor defines                      //
//////////////////////////////////////////////////////////

// Distantele minime la care ajunge
#define FRONT_MAXIM_VOLTAGE 410
#define BACK_MAXIM_VOLTAGE  200
#define RIGHT_MAXIM_VOLTAGE 500

// Voltajul maxim pentru care ce e in dreapta e considerat loc liber
#define RIGHT_FREE_SPACE_THRES_VOLTAGE 130

// Lungimea minima necesara pentru a intra dand o singura data in spate
#define MINIMUM_ONE_MOVE_PARK_LEN 6

// Lungimea minima a unui loc de parcare
#define MINIMUM_SPACE_LEN MINIMUM_ONE_MOVE_PARK_LEN

//////////////////////////////////////////////////////////
//                  Variabile globale                   //
//////////////////////////////////////////////////////////

// Ii spune timer-ului 1 sa gaseasca un loc de parcare
volatile int found_space = 0, find_space = 0, space_len = 0;

volatile int wait_ticks = 0, ticks_to_wait = 0;

// Indica daca a oprit masina in mod fortat pentru ca a aparut un obstacol
volatile int error_occured;

// Lungimea de masurat
volatile int measure_len = 0; 
volatile int measure_bad_go = 0;
// Lungimea pe care a parcurso in plus dupa ce a gasit locul de parcare
volatile int bad_go_len = 0;
// Indica intreruperii externe 2 sa decrementeze bad_go_len
volatile int decrement_bad_go = 0;


///////////////////////////////////////////////
// DEBUG

#define APRINDE_LED1    PORTA |= (1<<PA3)
#define APRINDE_LED2    PORTA |= (1<<PA5)
#define APRINDE_LED3    PORTA |= (1<<PA7)

#define STINGE_LED1     PORTA &= ~(1<<PA3)
#define STINGE_LED2     PORTA &= ~(1<<PA5)
#define STINGE_LED3     PORTA &= ~(1<<PA7)


#define APRINDE_LED4    PORTC |= (1<<PC1)
#define STINGE_LED4     PORTC &= ~(1<<PC1);


/************************************************************************ 
 * REDIRECTARE stdin, stdout pentru receptie, transmisie seriala
 ************************************************************************/
int uart_putchar(char c, FILE *unused);
int uart_getchar(FILE* f);
FILE mystdout = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);
FILE mystdin =  FDEV_SETUP_STREAM(NULL, uart_getchar, _FDEV_SETUP_READ);

//////////////////////////////////////////////////////////////////////////////////////
// Initializez interfata seriala
void USART_init(void)
{
	
	// Initializare modul USART pt comunicatie la 9600bps, 8 biti de date, la 16MHz											
	UCSRB = 1<<RXEN | 1<<TXEN;
	UCSRC = 1<<URSEL | 1<<UCSZ1|1<<UCSZ0;
	UBRRL = (char)(103&0xff);
	UBRRH = (char)(103>>8);
	
}

//////////////////////////////////////////////////////////////////////////////////////
// Face o conversie de pe channel si intoarce valoarea
int ADC_get(uint8_t channel)
{
	int rez;

	//setati canalul de pe care vreti sa faceti citirea
	ADMUX &= 0xE0;
	ADMUX |= channel;
	ADCSRA |= 1<<ADSC;
	
	//asteptati sa se termine o conversie
	while(!(ADCSRA & (1<<ADIF)))
		;
	
	//returnati rezultatul conversiei
	rez = ADCL;
	rez += (ADCH<<8);
	
	return rez;
}

//////////////////////////////////////////////////////////////////////////////////////
// Initializeaza ADC-ul
void ADC_init(void)
{
	//referinta AVCC cu condesator extern pe pinul AREF;
	ADMUX |= (1<<REFS0);// | (0<<REFS1);
	// activeaza ADC si il pune in modul free running (auto trigger). Prescaler de ceas la maximum. FARA intrerupere la conversie completa.
	ADCSRA |= (1<<ADEN) | (1<<ADATE) | (1<<ADPS0) | (1<<ADPS1) | (1<<ADPS2);
}

//////////////////////////////////////////////////////////////////////////////////////
// Initilizeaza timer 1 sa dea intrerupere la in jur de 20 ms
void Timer1_init(void)
{
	OCR1A = 300;
	TCCR1A = 0;
	TCCR1B = (1<<WGM12)|(1<<CS12)|(1<<CS10);
	TIMSK |= _BV(OCIE1A);
}

//////////////////////////////////////////////////////////////////////////////////////
// Functia apelata de timer1 la comparare 
SIGNAL(SIG_OUTPUT_COMPARE1A)
{	
	// Iau valoarea de la un senzor
	if (channel > 0)
		voltages[channel-1] = ADC_get(channels[channel]); //citim valoarea de la ADC
	else
		voltages[2] = ADC_get(channels[channel]); //citim valoarea de la ADC
	
	// Daca am luat de la toti trei senzorii deja:
	if (channel == BACK)
	{
		// Daca s-a cerut cautarea unui loc de parcare
		if (find_space)
		{
			// Daca sunt langa un loc liber
			if (measure_len == 1)
			{
				// S-a terminat locul liber
				if (voltages[RIGHT] > RIGHT_FREE_SPACE_THRES_VOLTAGE)
				{
					// Daca locul gasit este destul de lung
					if (space_len >= MINIMUM_SPACE_LEN)
					{
						// Anunt ca am gasit un loc
						APRINDE_LED3;
						
						bad_go_len = 0;
						measure_bad_go = 1;
						measure_len = 0;
						
						find_space = 0;
						found_space = 1;
					}
					// Daca e prea scurt locul resetez lungimea si caut altul
					else
					{
						APRINDE_LED1;
						measure_len = 0;
						space_len = 0;
					}
				}
			}
			// Daca inca nu a gasit loc
			else
			{
				// A detectat inceputul unui loc de parcare
				if (voltages[RIGHT] <= RIGHT_FREE_SPACE_THRES_VOLTAGE)
				{
					STINGE_LED1;
					STINGE_LED3;
					space_len = 0;
					measure_len = 1;
				}
			}
		}

		// Debug
		if (voltages[RIGHT] <= RIGHT_FREE_SPACE_THRES_VOLTAGE)
		{
			STINGE_LED4;
		}
		else
			APRINDE_LED4;
		
		/*
		if (front_back_status == GOING_BACKWARD && voltages[BACK] > BACK_MAXIM_VOLTAGE)
			{
				APRINDE_LED3;
				error_occured = 1;
				found_space = 1;
				wait_ticks = 0;
				
				go_back_till_collision = 0;
				STAY;
			}
		*/
	}
	
	channel = (channel + 1)%3;
	
}

//////////////////////////////////////////////////////////////////////////////////////
// Initializez motorul din fata
void BackMotor_init()
{
	// Porturile de comanda
	DDRD |= (1<<BACK_MOTOR_EN);
	DDRC |= (1<<BACK_MOTOR_IN_1) | (1<<BACK_MOTOR_IN_2);
	
	// Ii spun sa stea pe loc
	PORTC &= ~(1 << BACK_MOTOR_IN_1);
	PORTC &= ~(1 << BACK_MOTOR_IN_2);

	// Timerul 2
	OCR2 = DEFAULT_BACK_OCR;
	TCCR2 = 1<<WGM20 | 1<<COM21 | 1<<CS22;
}

//////////////////////////////////////////////////////////////////////////////////////	
// Initializez motorul din fata
void FrontMotor_init()
{
	// Portutile de comanda
	DDRD |= (1<<FRONT_MOTOR_IN_1) | (1<<FRONT_MOTOR_IN_2);
	DDRB |= (1<<FRONT_MOTOR_EN);
	
	// Ii spun sa nu intoarca rotile si enable sa fie mereu 1
	GO_STRAIGHT;
	PORTB |= (1<<FRONT_MOTOR_EN);
}

//////////////////////////////////////////////////////////////////////////////////////
// Initializeaza motoarele
void Motors_init()
{
	BackMotor_init();
	FrontMotor_init();
}

//////////////////////////////////////////////////////////////////////////////////////
// Atunci cand roata a facut jumatate de rotatie se apeleaza aceasta rutina
ISR(INT2_vect)
{
	// Debug
	PORTA ^= (1<<PA5);
	
	// Daca i s-a spus sa masoare lungimea unui loc de parcare
	if (measure_len == 1)
		space_len++;
		
	// Daca i s-a spus sa numere cat a mers masina in plus
	if (measure_bad_go == 1)
	{
		bad_go_len++;
	}
	
	// Daca i s-a spus sa mearga masina in urma si el sa semnaleze cand e gata
	if (decrement_bad_go)
	{
		if (bad_go_len > 0)
			bad_go_len--;
		else
			decrement_bad_go = 0;
	}
}

//////////////////////////////////////////////////////////////////////////////////////
// Initializez intreruperea 2
void INT2_init()
{
	GICR |= (1<<INT2);
}

//////////////////////////////////////////////////////////////////////////////////////
// Programul principal
int main(){
	int first_part;
	
	// Initializez seriala
	USART_init();
	stdin = &mystdin;
	stdout = &mystdout;
	
	// Initializez ADC-ul, timer1, motoarele, intreruperea 2
	ADC_init();
	Timer1_init();
	Motors_init();
	INT2_init();
	sei();
	
	// Initializez ledurile
	DDRC |= 1<<PC1;
	DDRA |= (1<<PA3) | (1<<PA5) | (1<<PA7);
	
	// Initializez butonul
	DDRD &= ~(1<<PD6);
	PORTD |= (1<<PD6);
	
	TURN_RIGHT;
	while (1);

	// Bucla infinita
	while (1)
	{
		// Astept sa apesi butonul
		while (PIND & (1<<PD6))
		{
			APRINDE_LED1;
			_delay_ms(500);
			STINGE_LED1;
			_delay_ms(500);
		}

		_delay_ms(500);

		// Detectez un loc de parcare
		found_space = 0;
		space_len = 0;
		GO_FORWARD;
		find_space = 1;
		while (!found_space);
	
		APRINDE_LED1;
	
		// Doar in caz de eroare se intampla asta
		if (space_len < MINIMUM_SPACE_LEN)
		{
			STAY;
			continue;
		}
		
		STAY;
	
		_delay_ms(1500);
	
		// Merg in urma cat am mers in plus
		measure_bad_go = 0;
		bad_go_len -= 3;
		if (bad_go_len > 0)
			decrement_bad_go = 1;
		else 
			bad_go_len = 0;
	
		GO_BACK;
	
		// Debug
		STINGE_LED1;
	
		while (decrement_bad_go);
	
		STAY;
	
		_delay_ms(500);
	
		APRINDE_LED3;
	
	
		// Intorc roata la dreapta si dau in spate
		space_len = MINIMUM_SPACE_LEN+2;
		first_part = space_len/2;
		bad_go_len = first_part;
	
		TURN_RIGHT;
	
		_delay_ms(1000);
		OCR2 = 110;
		_delay_ms(1000);
	
		GO_BACK;

		decrement_bad_go = 1;
	
		while (decrement_bad_go);
	
		STINGE_LED3;
	
		// Intorc roata la stanga si dau in spate
		STAY;
		GO_STRAIGHT;
		_delay_ms(1000);
		TURN_LEFT;
		_delay_ms(1000);
		GO_BACK;
	
		bad_go_len = space_len - first_part;
		decrement_bad_go = 1;
		while (decrement_bad_go);
	
	
		// Ma opresc
		GO_STRAIGHT;
		STAY;
		APRINDE_LED3;
	
	
		_delay_ms(1000);
		OCR2 = DEFAULT_BACK_OCR;
	
		_delay_ms(1000);
	
		// Parcarea e gata !!!!!
		// :D
	}
	

	return 0;
}

/************************************************************************ 
 * IMPLEMENTARE FUNCTII TRANSMISIE/RECEPTIE SERIALA, pentru mapare stdin, stdout
 ************************************************************************/
// Functie ce trateaza trimiterea unui caracter pe seriala
int uart_putchar(char c, FILE *unused)      
{
	if (c == '\n')
		uart_putchar('\r', 0);
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = c;
	return 0;
}

// Functie ce trateaza primirea unui caracter pe seriala
int uart_getchar(FILE* f)
{
	char c;
    
    loop_until_bit_is_set(UCSRA, RXC);
    c = UDR;
    if(c=='\r')
		c='\n';
    uart_putchar(c,NULL); 
    return c;
}